@@ -1,5 +1,4 @@
-from __future__ import print_function
-
+import logging
 import string
 import sys
 from collections import deque
@@ -8,17 +7,17 @@ import numpy as np
 from scipy.special import logsumexp
 from sklearn.base import BaseEstimator
 from sklearn.utils import check_array, check_random_state
-from sklearn.utils.validation import check_is_fitted
 
-from . import _hmmc
+from . import _hmmc, _utils
 from .utils import normalize, log_normalize, iter_from_X_lengths, log_mask_zero
 
 
+_log = logging.getLogger(__name__)
 #: Supported decoder algorithms.
 DECODER_ALGORITHMS = frozenset(("viterbi", "map"))
 
 
-class ConvergenceMonitor(object):
+class ConvergenceMonitor:
     """Monitors and reports convergence to :data:`sys.stderr`.
 
     Parameters
@@ -209,7 +208,7 @@ class _BaseHMM(BaseEstimator):
         # The stationary distribution is proportional to the left-eigenvector
         # associated with the largest eigenvalue (i.e., 1) of the transition
         # matrix.
-        check_is_fitted(self, "transmat_")
+        _utils.check_is_fitted(self, "transmat_")
         eigvals, eigvecs = np.linalg.eig(self.transmat_.T)
         eigvec = np.real_if_close(eigvecs[:, np.argmax(eigvals)])
         return eigvec / eigvec.sum()
@@ -239,7 +238,7 @@ class _BaseHMM(BaseEstimator):
         score : Compute the log probability under the model.
         decode : Find most likely state sequence corresponding to ``X``.
         """
-        check_is_fitted(self, "startprob_")
+        _utils.check_is_fitted(self, "startprob_")
         self._check()
 
         X = check_array(X)
@@ -278,7 +277,7 @@ class _BaseHMM(BaseEstimator):
             posteriors.
         decode : Find most likely state sequence corresponding to ``X``.
         """
-        check_is_fitted(self, "startprob_")
+        _utils.check_is_fitted(self, "startprob_")
         self._check()
 
         X = check_array(X)
@@ -331,7 +330,7 @@ class _BaseHMM(BaseEstimator):
             posteriors.
         score : Compute the log probability under the model.
         """
-        check_is_fitted(self, "startprob_")
+        _utils.check_is_fitted(self, "startprob_")
         self._check()
 
         algorithm = algorithm or self.algorithm
@@ -413,7 +412,7 @@ class _BaseHMM(BaseEstimator):
         state_sequence : array, shape (n_samples, )
             State sequence produced by the model.
         """
-        check_is_fitted(self, "startprob_")
+        _utils.check_is_fitted(self, "startprob_")
         self._check()
 
         if random_state is None:
@@ -485,6 +484,10 @@ class _BaseHMM(BaseEstimator):
             if self.monitor_.converged:
                 break
 
+        if (self.transmat_.sum(axis=1) == 0).any():
+            _log.warning("Some rows of transmat_ have zero sum because no "
+                         "transition from the state was ever observed.")
+
         return self
 
     def _do_viterbi_pass(self, framelogprob):
@@ -541,6 +544,12 @@ class _BaseHMM(BaseEstimator):
         if 't' in self.init_params or not hasattr(self, "transmat_"):
             self.transmat_ = np.full((self.n_components, self.n_components),
                                      init)
+        n_fit_scalars_per_param = self._get_n_fit_scalars_per_param()
+        n_fit_scalars = sum(n_fit_scalars_per_param[p] for p in self.params)
+        if X.size < n_fit_scalars:
+            _log.warning("Fitting a model with {} free scalar parameters with "
+                         "only {} data points will result in a degenerate "
+                         "solution.".format(n_fit_scalars, X.size))
 
     def _check(self):
         """Validates model parameters prior to fitting.
@@ -679,15 +688,18 @@ class _BaseHMM(BaseEstimator):
         stats : dict
             Sufficient statistics updated from all available samples.
         """
+        # If a prior is < 1, `prior - 1 + starts['start']` can be negative.  In
+        # that case maximization of (n1+e1) log p1 + ... + (ns+es) log ps under
+        # the conditions sum(p) = 1 and all(p >= 0) show that the negative
+        # terms can just be set to zero.
         # The ``np.where`` calls guard against updating forbidden states
         # or transitions in e.g. a left-right HMM.
         if 's' in self.params:
-            startprob_ = self.startprob_prior - 1.0 + stats['start']
-            self.startprob_ = np.where(self.startprob_ == 0.0,
-                                       self.startprob_, startprob_)
+            startprob_ = np.maximum(self.startprob_prior - 1 + stats['start'],
+                                    0)
+            self.startprob_ = np.where(self.startprob_ == 0, 0, startprob_)
             normalize(self.startprob_)
         if 't' in self.params:
-            transmat_ = self.transmat_prior - 1.0 + stats['trans']
-            self.transmat_ = np.where(self.transmat_ == 0.0,
-                                      self.transmat_, transmat_)
+            transmat_ = np.maximum(self.transmat_prior - 1 + stats['trans'], 0)
+            self.transmat_ = np.where(self.transmat_ == 0, 0, transmat_)
             normalize(self.transmat_, axis=1)
