@@ -10,6 +10,8 @@
 The :mod:`hmmlearn.hmm` module implements hidden Markov models.
 """
 
+import logging
+
 import numpy as np
 from scipy.special import logsumexp
 from sklearn import cluster
@@ -18,13 +20,24 @@ from sklearn.utils import check_random_state
 from . import _utils
 from .stats import log_multivariate_normal_density
 from .base import _BaseHMM
-from .utils import iter_from_X_lengths, normalize, fill_covars
+from .utils import (
+    fill_covars, iter_from_X_lengths, log_mask_zero, log_normalize, normalize)
 
 __all__ = ["GMMHMM", "GaussianHMM", "MultinomialHMM"]
 
+
+_log = logging.getLogger(__name__)
 COVARIANCE_TYPES = frozenset(("spherical", "diag", "full", "tied"))
 
 
+def _check_and_set_gaussian_n_features(model, X):
+    _, n_features = X.shape
+    if hasattr(model, "n_features") and model.n_features != n_features:
+        raise ValueError("Unexpected number of dimensions, got {} but "
+                         "expected {}".format(n_features, model.n_features))
+    model.n_features = n_features
+
+
 class GaussianHMM(_BaseHMM):
     r"""Hidden Markov Model with Gaussian emissions.
 
@@ -109,12 +122,12 @@ class GaussianHMM(_BaseHMM):
     monitor\_ : ConvergenceMonitor
         Monitor object used to check the convergence of EM.
 
-    transmat\_ : array, shape (n_components, n_components)
-        Matrix of transition probabilities between states.
-
     startprob\_ : array, shape (n_components, )
         Initial state occupation distribution.
 
+    transmat\_ : array, shape (n_components, n_components)
+        Matrix of transition probabilities between states.
+
     means\_ : array, shape (n_components, n_features)
         Mean parameters for each state.
 
@@ -124,15 +137,14 @@ class GaussianHMM(_BaseHMM):
         The shape depends on :attr:`covariance_type`::
 
             (n_components, )                        if "spherical",
-            (n_features, n_features)                if "tied",
             (n_components, n_features)              if "diag",
             (n_components, n_features, n_features)  if "full"
+            (n_features, n_features)                if "tied",
 
     Examples
     --------
     >>> from hmmlearn.hmm import GaussianHMM
-    >>> GaussianHMM(n_components=2)
-    ...                             #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
+    >>> GaussianHMM(n_components=2)  #doctest: +ELLIPSIS
     GaussianHMM(algorithm='viterbi',...
     """
     def __init__(self, n_components=1, covariance_type='diag',
@@ -165,10 +177,13 @@ class GaussianHMM(_BaseHMM):
 
     @covars_.setter
     def covars_(self, covars):
-        self._covars_ = np.asarray(covars).copy()
+        covars = np.array(covars, copy=True)
+        _utils._validate_covars(covars, self.covariance_type,
+                                self.n_components)
+        self._covars_ = covars
 
     def _check(self):
-        super(GaussianHMM, self)._check()
+        super()._check()
 
         self.means_ = np.asarray(self.means_)
         self.n_features = self.means_.shape[1]
@@ -177,18 +192,25 @@ class GaussianHMM(_BaseHMM):
             raise ValueError('covariance_type must be one of {}'
                              .format(COVARIANCE_TYPES))
 
-        _utils._validate_covars(self._covars_, self.covariance_type,
-                                self.n_components)
+    def _get_n_fit_scalars_per_param(self):
+        nc = self.n_components
+        nf = self.n_features
+        return {
+            "s": nc - 1,
+            "t": nc * (nc - 1),
+            "m": nc * nf,
+            "c": {
+                "spherical": nc,
+                "diag": nc * nf,
+                "full": nc * nf * (nf + 1) // 2,
+                "tied": nf * (nf + 1) // 2,
+            }[self.covariance_type],
+        }
 
     def _init(self, X, lengths=None):
-        super(GaussianHMM, self)._init(X, lengths=lengths)
+        _check_and_set_gaussian_n_features(self, X)
+        super()._init(X, lengths=lengths)
 
-        _, n_features = X.shape
-        if hasattr(self, 'n_features') and self.n_features != n_features:
-            raise ValueError('Unexpected number of dimensions, got %s but '
-                             'expected %s' % (n_features, self.n_features))
-
-        self.n_features = n_features
         if 'm' in self.init_params or not hasattr(self, "means_"):
             kmeans = cluster.KMeans(n_clusters=self.n_components,
                                     random_state=self.random_state)
@@ -198,7 +220,7 @@ class GaussianHMM(_BaseHMM):
             cv = np.cov(X.T) + self.min_covar * np.eye(X.shape[1])
             if not cv.shape:
                 cv.shape = (1, 1)
-            self._covars_ = \
+            self.covars_ = \
                 _utils.distribute_covar_matrix_to_match_covariance_type(
                     cv, self.covariance_type, self.n_components).copy()
 
@@ -213,7 +235,7 @@ class GaussianHMM(_BaseHMM):
         )
 
     def _initialize_sufficient_statistics(self):
-        stats = super(GaussianHMM, self)._initialize_sufficient_statistics()
+        stats = super()._initialize_sufficient_statistics()
         stats['post'] = np.zeros(self.n_components)
         stats['obs'] = np.zeros((self.n_components, self.n_features))
         stats['obs**2'] = np.zeros((self.n_components, self.n_features))
@@ -224,7 +246,7 @@ class GaussianHMM(_BaseHMM):
 
     def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                           posteriors, fwdlattice, bwdlattice):
-        super(GaussianHMM, self)._accumulate_sufficient_statistics(
+        super()._accumulate_sufficient_statistics(
             stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice)
 
         if 'm' in self.params or 'c' in self.params:
@@ -241,7 +263,7 @@ class GaussianHMM(_BaseHMM):
                     'ij,ik,il->jkl', posteriors, obs, obs)
 
     def _do_mstep(self, stats):
-        super(GaussianHMM, self)._do_mstep(stats)
+        super()._do_mstep(stats)
 
         means_prior = self.means_prior
         means_weight = self.means_weight
@@ -349,20 +371,19 @@ class MultinomialHMM(_BaseHMM):
     monitor\_ : ConvergenceMonitor
         Monitor object used to check the convergence of EM.
 
-    transmat\_ : array, shape (n_components, n_components)
-        Matrix of transition probabilities between states.
-
     startprob\_ : array, shape (n_components, )
         Initial state occupation distribution.
 
+    transmat\_ : array, shape (n_components, n_components)
+        Matrix of transition probabilities between states.
+
     emissionprob\_ : array, shape (n_components, n_features)
         Probability of emitting a given symbol when in each state.
 
     Examples
     --------
     >>> from hmmlearn.hmm import MultinomialHMM
-    >>> MultinomialHMM(n_components=2)
-    ...                             #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
+    >>> MultinomialHMM(n_components=2)  #doctest: +ELLIPSIS
     MultinomialHMM(algorithm='viterbi',...
     """
     # TODO: accept the prior on emissionprob_ for consistency.
@@ -379,26 +400,27 @@ class MultinomialHMM(_BaseHMM):
                           n_iter=n_iter, tol=tol, verbose=verbose,
                           params=params, init_params=init_params)
 
-    def _init(self, X, lengths=None):
-        if not self._check_input_symbols(X):
-            raise ValueError("expected a sample from "
-                             "a Multinomial distribution.")
+    def _get_n_fit_scalars_per_param(self):
+        nc = self.n_components
+        nf = self.n_features
+        return {
+            "s": nc - 1,
+            "t": nc * (nc - 1),
+            "e": nc * (nf - 1),
+        }
 
-        super(MultinomialHMM, self)._init(X, lengths=lengths)
+    def _init(self, X, lengths=None):
+        self._check_and_set_n_features(X)
+        super()._init(X, lengths=lengths)
         self.random_state = check_random_state(self.random_state)
 
         if 'e' in self.init_params:
-            if not hasattr(self, "n_features"):
-                symbols = set()
-                for i, j in iter_from_X_lengths(X, lengths):
-                    symbols |= set(X[i:j].flatten())
-                self.n_features = len(symbols)
             self.emissionprob_ = self.random_state \
                 .rand(self.n_components, self.n_features)
             normalize(self.emissionprob_, axis=1)
 
     def _check(self):
-        super(MultinomialHMM, self)._check()
+        super()._check()
 
         self.emissionprob_ = np.atleast_2d(self.emissionprob_)
         n_features = getattr(self, "n_features", self.emissionprob_.shape[1])
@@ -409,7 +431,7 @@ class MultinomialHMM(_BaseHMM):
             self.n_features = n_features
 
     def _compute_log_likelihood(self, X):
-        return np.log(self.emissionprob_)[:, np.concatenate(X)].T
+        return log_mask_zero(self.emissionprob_)[:, np.concatenate(X)].T
 
     def _generate_sample_from_state(self, state, random_state=None):
         cdf = np.cumsum(self.emissionprob_[state, :])
@@ -417,41 +439,40 @@ class MultinomialHMM(_BaseHMM):
         return [(cdf > random_state.rand()).argmax()]
 
     def _initialize_sufficient_statistics(self):
-        stats = super(MultinomialHMM, self)._initialize_sufficient_statistics()
+        stats = super()._initialize_sufficient_statistics()
         stats['obs'] = np.zeros((self.n_components, self.n_features))
         return stats
 
     def _accumulate_sufficient_statistics(self, stats, X, framelogprob,
                                           posteriors, fwdlattice, bwdlattice):
-        super(MultinomialHMM, self)._accumulate_sufficient_statistics(
+        super()._accumulate_sufficient_statistics(
             stats, X, framelogprob, posteriors, fwdlattice, bwdlattice)
         if 'e' in self.params:
             for t, symbol in enumerate(np.concatenate(X)):
                 stats['obs'][:, symbol] += posteriors[t]
 
     def _do_mstep(self, stats):
-        super(MultinomialHMM, self)._do_mstep(stats)
+        super()._do_mstep(stats)
         if 'e' in self.params:
             self.emissionprob_ = (stats['obs']
                                   / stats['obs'].sum(axis=1)[:, np.newaxis])
 
-    def _check_input_symbols(self, X):
-        """Check if ``X`` is a sample from a Multinomial distribution.
-
-        That is ``X`` should be an array of non-negative integers from
-        range ``[min(X), max(X)]``, such that each integer from the range
-        occurs in ``X`` at least once.
-
-        For example ``[0, 0, 2, 1, 3, 1, 1]`` is a valid sample from a
-        Multinomial distribution, while ``[0, 0, 3, 5, 10]`` is not.
+    def _check_and_set_n_features(self, X):
         """
-        symbols = np.concatenate(X)
-        if (len(symbols) == 1                                # not enough data
-            or not np.issubdtype(symbols.dtype, np.integer)  # not an integer
-            or (symbols < 0).any()):                         # not positive
-            return False
-        u = np.unique(symbols)
-        return u[0] == 0 and u[-1] == len(u) - 1
+        Check if ``X`` is a sample from a Multinomial distribution, i.e. an
+        array of non-negative integers.
+        """
+        if not np.issubdtype(X.dtype, np.integer):
+            raise ValueError("Symbols should be integers")
+        if X.min() < 0:
+            raise ValueError("Symbols should be nonnegative")
+        if hasattr(self, "n_features"):
+            if self.n_features - 1 < X.max():
+                raise ValueError(
+                    "Largest symbol is {} but the model only emits "
+                    "symbols up to {}"
+                    .format(X.max(), self.n_features - 1))
+        self.n_features = X.max() + 1
 
 
 class GMMHMM(_BaseHMM):
@@ -474,7 +495,9 @@ class GMMHMM(_BaseHMM):
         * "diag" --- each state uses a diagonal covariance matrix.
         * "full" --- each state uses a full (i.e. unrestricted)
           covariance matrix.
-        * "tied" --- all states use **the same** full covariance matrix.
+        * "tied" --- all mixture components of each state use **the same** full
+          covariance matrix (note that this is not the same as for
+          `GaussianHMM`).
 
         Defaults to "diag".
 
@@ -560,9 +583,9 @@ class GMMHMM(_BaseHMM):
         The shape depends on :attr:`covariance_type`::
 
             (n_components, n_mix)                          if "spherical",
-            (n_components, n_features, n_features)         if "tied",
             (n_components, n_mix, n_features)              if "diag",
             (n_components, n_mix, n_features, n_features)  if "full"
+            (n_components, n_features, n_features)         if "tied",
     """
 
     def __init__(self, n_components=1, n_mix=1,
@@ -588,54 +611,66 @@ class GMMHMM(_BaseHMM):
         self.covars_prior = covars_prior
         self.covars_weight = covars_weight
 
-    def _init(self, X, lengths=None):
-        super(GMMHMM, self)._init(X, lengths=lengths)
+    def _get_n_fit_scalars_per_param(self):
+        nc = self.n_components
+        nf = self.n_features
+        nm = self.n_mix
+        return {
+            "s": nc - 1,
+            "t": nc * (nc - 1),
+            "m": nc * nm * nf,
+            "c": {
+                "spherical": nc * nm,
+                "diag": nc * nm * nf,
+                "full": nc * nm * nf * (nf + 1) // 2,
+                "tied": nc * nf * (nf + 1) // 2,
+            }[self.covariance_type],
+            "w": nm - 1,
+        }
 
-        _n_samples, self.n_features = X.shape
+    def _init(self, X, lengths=None):
+        _check_and_set_gaussian_n_features(self, X)
+        super()._init(X, lengths=lengths)
+        nc = self.n_components
+        nf = self.n_features
+        nm = self.n_mix
 
         # Default values for covariance prior parameters
         self._init_covar_priors()
         self._fix_priors_shape()
 
-        main_kmeans = cluster.KMeans(n_clusters=self.n_components,
+        main_kmeans = cluster.KMeans(n_clusters=nc,
                                      random_state=self.random_state)
         labels = main_kmeans.fit_predict(X)
         kmeanses = []
-        for label in range(self.n_components):
-            kmeans = cluster.KMeans(n_clusters=self.n_mix,
+        for label in range(nc):
+            kmeans = cluster.KMeans(n_clusters=nm,
                                     random_state=self.random_state)
             kmeans.fit(X[np.where(labels == label)])
             kmeanses.append(kmeans)
 
         if 'w' in self.init_params or not hasattr(self, "weights_"):
-            self.weights_ = (np.ones((self.n_components, self.n_mix)) /
-                             (np.ones((self.n_components, 1)) * self.n_mix))
+            self.weights_ = np.ones((nc, nm)) / (np.ones((nc, 1)) * nm)
 
         if 'm' in self.init_params or not hasattr(self, "means_"):
-            self.means_ = np.zeros((self.n_components, self.n_mix,
-                                    self.n_features))
-            for i, kmeans in enumerate(kmeanses):
-                self.means_[i] = kmeans.cluster_centers_
+            self.means_ = np.stack(
+                [kmeans.cluster_centers_ for kmeans in kmeanses])
 
         if 'c' in self.init_params or not hasattr(self, "covars_"):
-            cv = np.cov(X.T) + self.min_covar * np.eye(self.n_features)
+            cv = np.cov(X.T) + self.min_covar * np.eye(nf)
             if not cv.shape:
                 cv.shape = (1, 1)
-
             if self.covariance_type == 'tied':
-                self.covars_ = np.zeros((self.n_components,
-                                         self.n_features, self.n_features))
+                self.covars_ = np.zeros((nc, nf, nf))
                 self.covars_[:] = cv
             elif self.covariance_type == 'full':
-                self.covars_ = np.zeros((self.n_components, self.n_mix,
-                                         self.n_features, self.n_features))
+                self.covars_ = np.zeros((nc, nm, nf, nf))
                 self.covars_[:] = cv
             elif self.covariance_type == 'diag':
-                self.covars_ = np.zeros((self.n_components, self.n_mix,
-                                         self.n_features))
+                self.covars_ = np.zeros((nc, nm, nf))
                 self.covars_[:] = np.diag(cv)
             elif self.covariance_type == 'spherical':
-                self.covars_ = np.zeros((self.n_components, self.n_mix))
+                self.covars_ = np.zeros((nc, nm))
                 self.covars_[:] = cv.mean()
 
     def _init_covar_priors(self):
@@ -661,48 +696,47 @@ class GMMHMM(_BaseHMM):
                 self.covars_weight = 0.0
 
     def _fix_priors_shape(self):
+        nc = self.n_components
+        nf = self.n_features
+        nm = self.n_mix
+
         # If priors are numbers, this function will make them into a
         # matrix of proper shape
         self.weights_prior = np.broadcast_to(
-            self.weights_prior, (self.n_components, self.n_mix)).copy()
+            self.weights_prior, (nc, nm)).copy()
         self.means_prior = np.broadcast_to(
-            self.means_prior,
-            (self.n_components, self.n_mix, self.n_features)).copy()
+            self.means_prior, (nc, nm, nf)).copy()
         self.means_weight = np.broadcast_to(
-            self.means_weight,
-            (self.n_components, self.n_mix)).copy()
+            self.means_weight, (nc, nm)).copy()
 
         if self.covariance_type == "full":
             self.covars_prior = np.broadcast_to(
-                self.covars_prior,
-                (self.n_components, self.n_mix,
-                 self.n_features, self.n_features)).copy()
+                self.covars_prior, (nc, nm, nf, nf)).copy()
             self.covars_weight = np.broadcast_to(
-                self.covars_weight, (self.n_components, self.n_mix)).copy()
+                self.covars_weight, (nc, nm)).copy()
         elif self.covariance_type == "tied":
             self.covars_prior = np.broadcast_to(
-                self.covars_prior,
-                (self.n_components, self.n_features, self.n_features)).copy()
+                self.covars_prior, (nc, nf, nf)).copy()
             self.covars_weight = np.broadcast_to(
-                self.covars_weight, self.n_components).copy()
+                self.covars_weight, nc).copy()
         elif self.covariance_type == "diag":
             self.covars_prior = np.broadcast_to(
-                self.covars_prior,
-                (self.n_components, self.n_mix, self.n_features)).copy()
+                self.covars_prior, (nc, nm, nf)).copy()
             self.covars_weight = np.broadcast_to(
-                self.covars_weight,
-                (self.n_components, self.n_mix, self.n_features)).copy()
+                self.covars_weight, (nc, nm, nf)).copy()
         elif self.covariance_type == "spherical":
             self.covars_prior = np.broadcast_to(
-                self.covars_prior, (self.n_components, self.n_mix)).copy()
+                self.covars_prior, (nc, nm)).copy()
             self.covars_weight = np.broadcast_to(
-                self.covars_weight, (self.n_components, self.n_mix)).copy()
+                self.covars_weight, (nc, nm)).copy()
 
     def _check(self):
-        super(GMMHMM, self)._check()
-
+        super()._check()
         if not hasattr(self, "n_features"):
             self.n_features = self.means_.shape[2]
+        nc = self.n_components
+        nf = self.n_features
+        nm = self.n_mix
 
         self._init_covar_priors()
         self._fix_priors_shape()
@@ -714,20 +748,18 @@ class GMMHMM(_BaseHMM):
 
         self.weights_ = np.array(self.weights_)
         # Checking mixture weights' shape
-        if self.weights_.shape != (self.n_components, self.n_mix):
+        if self.weights_.shape != (nc, nm):
             raise ValueError("mixture weights must have shape "
                              "(n_components, n_mix), actual shape: {}"
                              .format(self.weights_.shape))
 
         # Checking mixture weights' mathematical correctness
-        if not np.allclose(np.sum(self.weights_, axis=1),
-                           np.ones(self.n_components)):
+        if not np.allclose(np.sum(self.weights_, axis=1), np.ones(nc)):
             raise ValueError("mixture weights must sum up to 1")
 
         # Checking means' shape
         self.means_ = np.array(self.means_)
-        if self.means_.shape != (self.n_components, self.n_mix,
-                                 self.n_features):
+        if self.means_.shape != (nc, nm, nf):
             raise ValueError("mixture means must have shape "
                              "(n_components, n_mix, n_features), "
                              "actual shape: {}".format(self.means_.shape))
@@ -736,11 +768,10 @@ class GMMHMM(_BaseHMM):
         self.covars_ = np.array(self.covars_)
         covars_shape = self.covars_.shape
         needed_shapes = {
-            "spherical": (self.n_components, self.n_mix),
-            "tied": (self.n_components, self.n_features, self.n_features),
-            "diag": (self.n_components, self.n_mix, self.n_features),
-            "full": (self.n_components, self.n_mix,
-                     self.n_features, self.n_features)
+            "spherical": (nc, nm),
+            "tied": (nc, nf, nf),
+            "diag": (nc, nm, nf),
+            "full": (nc, nm, nf, nf),
         }
         needed_shape = needed_shapes[self.covariance_type]
         if covars_shape != needed_shape:
@@ -754,24 +785,38 @@ class GMMHMM(_BaseHMM):
 
         if (self.covariance_type == "spherical" or
                 self.covariance_type == "diag"):
-            if np.any(self.covars_ <= 0):
+            if np.any(self.covars_ < 0):
                 raise ValueError("{!r} mixture covars must be non-negative"
                                  .format(self.covariance_type))
+            if np.any(self.covars_ == 0):
+                _log.warning("Degenerate mixture covariance")
         elif self.covariance_type == "tied":
             for i, covar in enumerate(self.covars_):
-                if (not np.allclose(covar, covar.T) or
-                        np.any(linalg.eigvalsh(covar) <= 0)):
-                    raise ValueError("'tied' mixture covars must be "
-                                     "symmetric, positive-definite")
+                if not np.allclose(covar, covar.T):
+                    raise ValueError("Covariance of state #{} is not symmetric"
+                                     .format(i))
+                min_eigvalsh = np.linalg.eigvalsh(covar).min()
+                if min_eigvalsh < 0:
+                    raise ValueError("Covariance of state #{} is not positive "
+                                     "definite".format(i))
+                if min_eigvalsh == 0:
+                    _log.warning("Covariance of state #%d has a null "
+                                 "eigenvalue.", i)
         elif self.covariance_type == "full":
             for i, mix_covars in enumerate(self.covars_):
                 for j, covar in enumerate(mix_covars):
-                    if (not np.allclose(covar, covar.T) or
-                            np.any(linalg.eigvalsh(covar) <= 0)):
+                    if not np.allclose(covar, covar.T):
                         raise ValueError(
-                            "'full' covariance matrix of mixture {} of "
-                            "component {} must be symmetric, positive-definite"
-                            .format(j, i))
+                            "Covariance of state #{}, mixture #{} is not "
+                            "symmetric".format(i, j))
+                    min_eigvalsh = np.linalg.eigvalsh(covar).min()
+                    if min_eigvalsh < 0:
+                        raise ValueError(
+                            "Covariance of state #{}, mixture #{} is not "
+                            "positive definite".format(i, j))
+                    if min_eigvalsh == 0:
+                        _log.warning("Covariance of state #%d, mixture #%d "
+                                     "has a null eigenvalue.", i, j)
 
     def _generate_sample_from_state(self, state, random_state=None):
         if random_state is None:
@@ -815,7 +860,7 @@ class GMMHMM(_BaseHMM):
         return res
 
     def _initialize_sufficient_statistics(self):
-        stats = super(GMMHMM, self)._initialize_sufficient_statistics()
+        stats = super()._initialize_sufficient_statistics()
         stats['n_samples'] = 0
         stats['post_comp_mix'] = None
         stats['post_mix_sum'] = np.zeros((self.n_components, self.n_mix))
@@ -829,7 +874,7 @@ class GMMHMM(_BaseHMM):
 
         # TODO: support multiple frames
 
-        super(GMMHMM, self)._accumulate_sufficient_statistics(
+        super()._accumulate_sufficient_statistics(
             stats, X, framelogprob, post_comp, fwdlattice, bwdlattice
         )
 
@@ -838,15 +883,15 @@ class GMMHMM(_BaseHMM):
         stats['n_samples'] = n_samples
         stats['samples'] = X
 
-        prob_mix = np.zeros((n_samples, self.n_components, self.n_mix))
+        post_mix = np.zeros((n_samples, self.n_components, self.n_mix))
         for p in range(self.n_components):
             log_denses = self._compute_log_weighted_gaussian_densities(X, p)
+            log_normalize(log_denses, axis=-1)
             with np.errstate(under="ignore"):
-                prob_mix[:, p, :] = np.exp(log_denses) + np.finfo(np.float).eps
+                post_mix[:, p, :] = np.exp(log_denses)
 
-        prob_mix_sum = np.sum(prob_mix, axis=2)
-        post_mix = prob_mix / prob_mix_sum[:, :, np.newaxis]
-        post_comp_mix = post_comp[:, :, np.newaxis] * post_mix
+        with np.errstate(under="ignore"):
+            post_comp_mix = post_comp[:, :, np.newaxis] * post_mix
         stats['post_comp_mix'] = post_comp_mix
 
         stats['post_mix_sum'] = np.sum(post_comp_mix, axis=0)
@@ -855,10 +900,12 @@ class GMMHMM(_BaseHMM):
         stats['centered'] = X[:, np.newaxis, np.newaxis, :] - self.means_
 
     def _do_mstep(self, stats):
-        super(GMMHMM, self)._do_mstep(stats)
+        super()._do_mstep(stats)
+        nc = self.n_components
+        nf = self.n_features
+        nm = self.n_mix
 
         n_samples = stats['n_samples']
-        n_features = self.n_features
 
         # Maximizing weights
         alphas_minus_one = self.weights_prior - 1
@@ -870,10 +917,10 @@ class GMMHMM(_BaseHMM):
 
         # Maximizing means
         lambdas, mus = self.means_weight, self.means_prior
-        new_means_numer = np.einsum(
-            'ijk,il->jkl',
-            stats['post_comp_mix'], stats['samples']
-        ) + lambdas[:, :, np.newaxis] * mus
+        new_means_numer = (
+            np.einsum('ijk,il->jkl', stats['post_comp_mix'], stats['samples'])
+            + lambdas[:, :, np.newaxis] * mus
+        )
         new_means_denom = (stats['post_mix_sum'] + lambdas)[:, :, np.newaxis]
         new_means = new_means_numer / new_means_denom
 
@@ -881,35 +928,28 @@ class GMMHMM(_BaseHMM):
         centered_means = self.means_ - mus
 
         if self.covariance_type == 'full':
-            centered = stats['centered'].reshape((
-                n_samples, self.n_components, self.n_mix, self.n_features, 1
-            ))
-            centered_t = stats['centered'].reshape((
-                n_samples, self.n_components, self.n_mix, 1, self.n_features
-            ))
+            centered = stats['centered'].reshape((n_samples, nc, nm, nf, 1))
+            centered_t = stats['centered'].reshape((n_samples, nc, nm, 1, nf))
             centered_dots = centered * centered_t
 
             psis_t = np.transpose(self.covars_prior, axes=(0, 1, 3, 2))
             nus = self.covars_weight
 
-            centr_means_resh = centered_means.reshape((
-                self.n_components, self.n_mix, self.n_features, 1
-            ))
-            centr_means_resh_t = centered_means.reshape((
-                self.n_components, self.n_mix, 1, self.n_features
-            ))
+            centr_means_resh = centered_means.reshape((nc, nm, nf, 1))
+            centr_means_resh_t = centered_means.reshape((nc, nm, 1, nf))
             centered_means_dots = centr_means_resh * centr_means_resh_t
 
-            new_cov_numer = np.einsum(
-                'ijk,ijklm->jklm',
-                stats['post_comp_mix'], centered_dots
-            ) + psis_t + (lambdas[:, :, np.newaxis, np.newaxis] *
-                          centered_means_dots)
+            new_cov_numer = (
+                np.einsum(
+                    'ijk,ijklm->jklm', stats['post_comp_mix'], centered_dots)
+                + psis_t
+                + lambdas[:, :, np.newaxis, np.newaxis] * centered_means_dots
+            )
             new_cov_denom = (
-                stats['post_mix_sum'] + 1 + nus + self.n_features + 1
+                stats['post_mix_sum'] + 1 + nus + nf + 1
             )[:, :, np.newaxis, np.newaxis]
-
             new_cov = new_cov_numer / new_cov_denom
+
         elif self.covariance_type == 'diag':
             centered2 = stats['centered'] ** 2
             centered_means2 = centered_means ** 2
@@ -917,15 +957,16 @@ class GMMHMM(_BaseHMM):
             alphas = self.covars_prior
             betas = self.covars_weight
 
-            new_cov_numer = np.einsum(
-                'ijk,ijkl->jkl',
-                stats['post_comp_mix'], centered2
-            ) + lambdas[:, :, np.newaxis] * centered_means2 + 2 * betas
+            new_cov_numer = (
+                np.einsum('ijk,ijkl->jkl', stats['post_comp_mix'], centered2)
+                + lambdas[:, :, np.newaxis] * centered_means2
+                + 2 * betas
+            )
             new_cov_denom = (
                 stats['post_mix_sum'][:, :, np.newaxis] + 1 + 2 * (alphas + 1)
             )
-
             new_cov = new_cov_numer / new_cov_denom
+
         elif self.covariance_type == 'spherical':
             centered_norm2 = np.sum(stats['centered'] ** 2, axis=-1)
 
@@ -934,49 +975,37 @@ class GMMHMM(_BaseHMM):
 
             centered_means_norm2 = np.sum(centered_means ** 2, axis=-1)
 
-            new_cov_numer = np.einsum(
-                'ijk,ijk->jk',
-                stats['post_comp_mix'], centered_norm2
-            ) + lambdas * centered_means_norm2 + 2 * betas
-            new_cov_denom = (
-                n_features * stats['post_mix_sum'] + n_features +
-                2 * (alphas + 1)
+            new_cov_numer = (
+                np.einsum(
+                    'ijk,ijk->jk', stats['post_comp_mix'], centered_norm2)
+                + lambdas * centered_means_norm2
+                + 2 * betas
             )
-
+            new_cov_denom = nf * (stats['post_mix_sum'] + 1) + 2 * (alphas + 1)
             new_cov = new_cov_numer / new_cov_denom
+
         elif self.covariance_type == 'tied':
-            centered = stats['centered'].reshape((
-                n_samples, self.n_components, self.n_mix, self.n_features, 1
-            ))
-            centered_t = stats['centered'].reshape((
-                n_samples, self.n_components, self.n_mix, 1, self.n_features
-            ))
+            centered = stats['centered'].reshape((n_samples, nc, nm, nf, 1))
+            centered_t = stats['centered'].reshape((n_samples, nc, nm, 1, nf))
             centered_dots = centered * centered_t
 
             psis_t = np.transpose(self.covars_prior, axes=(0, 2, 1))
             nus = self.covars_weight
 
-            centr_means_resh = centered_means.reshape((
-                self.n_components, self.n_mix, self.n_features, 1
-            ))
-            centr_means_resh_t = centered_means.reshape((
-                self.n_components, self.n_mix, 1, self.n_features
-            ))
+            centr_means_resh = centered_means.reshape((nc, nm, nf, 1))
+            centr_means_resh_t = centered_means.reshape((nc, nm, 1, nf))
             centered_means_dots = centr_means_resh * centr_means_resh_t
 
-            lambdas_cmdots_prod_sum = np.einsum(
-                'ij,ijkl->ikl',
-                lambdas, centered_means_dots
-            )
+            lambdas_cmdots_prod_sum = (
+                np.einsum('ij,ijkl->ikl', lambdas, centered_means_dots))
 
-            new_cov_numer = np.einsum(
-                'ijk,ijklm->jlm',
-                stats['post_comp_mix'], centered_dots
-            ) + lambdas_cmdots_prod_sum + psis_t
+            new_cov_numer = (
+                np.einsum(
+                    'ijk,ijklm->jlm', stats['post_comp_mix'], centered_dots)
+                + lambdas_cmdots_prod_sum + psis_t)
             new_cov_denom = (
-                stats['post_sum'] + self.n_mix + nus + self.n_features + 1
+                stats['post_sum'] + nm + nus + nf + 1
             )[:, np.newaxis, np.newaxis]
-
             new_cov = new_cov_numer / new_cov_denom
 
         # Assigning new values to class members
